'From Squeak 3.2 of 11 July 2002 [latest update: #4917] on 29 July 2002 at 10:58:10 pm'!
"Change Set:		MaClientServerSupport
Date:			29 July 2002
Author:			Chris Muller

Includes a socket facade, called MaSynchronousSocket that works with MaSynchronousServerSocket to take care of archaic socket details that are unnecessary for syncronous client-server communications.

Also, a general, proprietary distributed computing framework.  Utilizes MaObjectSerialization to objectify the protocol into a hierarchy of classes defined by you.

To set up the server, create a requestProcessor class that implements the methods required by MaRequestServer.  Then:

	(MaRequestServerForTcp requestProcessor: MyRequestProcessorClass new)
		processRequestsOn: <portNumber>
		until: [ <boolean expression> ]"!

MaObject subclass: #MaByteArraySpecification
	instanceVariableNames: 'byteArray startPosition numberOfBytes '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Network-Client server'!

!MaByteArraySpecification commentStamp: '<historical>' prior: 0!
I merely hold a ByteArray and other information about it useful for efficient socket communications.!

MaObject subclass: #MaRequestServer
	instanceVariableNames: 'requestProcessor '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Network-Client server base'!
MaRequestServer subclass: #MaLocalRequestServer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Network-Client server'!

!MaLocalRequestServer commentStamp: '<historical>' prior: 0!
No local-specific behavior at this time.  Currently, it exists only to semantically counter MaRemoteRequestServer.!

MaRequestServer subclass: #MaRequestServerForTcp
	instanceVariableNames: 'socket serializer '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Network-Client server'!
MaRequestServerForTcp class
	instanceVariableNames: ''!
MaObject subclass: #MaServerLink
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Network-Client server private'!
MaServerLink subclass: #MaLocalServerLink
	instanceVariableNames: 'localRequestServer '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Network-Client server private'!
MaServerLink subclass: #MaRemoteServerLink
	instanceVariableNames: 'serializer socket '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Network-Client server private'!
MaObject subclass: #MaServerRequest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Network-Client server base'!

!MaServerRequest commentStamp: '<historical>' prior: 0!
It is not required to use this class.  It is only to demonstrate that you may wish to objectify your requests.!

MaObject subclass: #MaSynchronousSocket
	instanceVariableNames: 'hostAddress port socket '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Network-Sockets'!

!MaSynchronousSocket commentStamp: '<historical>' prior: 0!
I provide a means to communicate with a another running Squeak image, a "server" image, that is using MaSynchronousServerSocket to listen.  Data is sent and received using my 'send' methods.

To create an instance of me, see my class-side 'creation' methods.  If you use #new instead, be sure you set the hostAddress: and port: the other image is listening on before you try to send anything or, obviously, I will have to throw a UserError.

You can create and maintain an instance whether the server image is listening or not, but it must be listening to use the send method(s) since that's the time it actually establishes the connection.  Data is immediately sent and I wait for the response eventually returned to you.  See my 'configuration' methods to adjust my timeout values.




Other notes:
It is not recommended to store sockets in Magma databases.!

MaSynchronousSocket subclass: #MaSynchronousServerSocket
	instanceVariableNames: 'requestSpecification responseSpecification '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Network-Sockets'!

!MaByteArraySpecification methodsFor: 'converting' stamp: 'cmm 6/12/2002 23:41'!
logicalByteArray

	^byteArray
		copyFrom: startPosition
		to: startPosition + numberOfBytes - 1
! !

!MaByteArraySpecification methodsFor: 'accessing' stamp: 'cmm 6/11/2002 23:48'!
byteArray

	^byteArray
! !

!MaByteArraySpecification methodsFor: 'accessing' stamp: 'cmm 6/11/2002 23:48'!
byteArray: aByteArray

	byteArray _ aByteArray! !

!MaByteArraySpecification methodsFor: 'accessing' stamp: 'cmm 6/12/2002 00:12'!
numberOfBytes

	^numberOfBytes isNil
		ifTrue: [ byteArray size ]
		ifFalse: [ numberOfBytes ]
! !

!MaByteArraySpecification methodsFor: 'accessing' stamp: 'cmm 6/11/2002 23:58'!
numberOfBytes: anInteger

	numberOfBytes _ anInteger
! !

!MaByteArraySpecification methodsFor: 'accessing' stamp: 'cmm 6/11/2002 23:48'!
startPosition

	^startPosition! !

!MaByteArraySpecification methodsFor: 'accessing' stamp: 'cmm 6/11/2002 23:48'!
startPosition: anInteger

	startPosition _ anInteger! !

!MaByteArraySpecification methodsFor: 'accessing' stamp: 'cmm 6/12/2002 00:07'!
useAllBytes

	numberOfBytes _ nil! !

!MaByteArraySpecification methodsFor: 'initializing' stamp: 'cmm 6/12/2002 23:49'!
initialize

	startPosition _ 1.
	self useAllBytes
! !


!MaByteArraySpecification class methodsFor: 'as yet unclassified' stamp: 'cmm 6/12/2002 23:38'!
new: anInteger

	^self new
		byteArray: (ByteArray new: anInteger);
		yourself! !


!MaRequestServer methodsFor: 'initializing' stamp: 'cmm 4/1/2002 00:51'!
requestProcessor: processor
	"processor is a class that responds to #processRequest:"

	requestProcessor _ processor
! !

!MaRequestServer methodsFor: 'private' stamp: 'cmm 6/2/2002 21:38'!
processRequest: aMaServerRequest
	"Here's your request object, do something with it and answer any object, though every class throughout the entire responses graph must be part of your protocol or clients will be unable to materialize them."

Transcript cr; show: 'processing ', aMaServerRequest printString.
	^requestProcessor processRequest: aMaServerRequest
! !


!MaRequestServer class methodsFor: 'creation' stamp: 'cmm 3/31/2002 22:57'!
requestProcessor: requestProcessor
	"requestProcessor -> a class that implements processRequest: <one of your request objects>"

	^self new
		requestProcessor: requestProcessor;
		yourself
! !


!MaRequestServerForTcp methodsFor: 'listen' stamp: 'cmm 6/13/2002 23:16'!
processRequestsOn: portNumber until: aBlock

	socket
		listenOn: portNumber
		do:
			[ :requestSpec :responseSpec |
			self
				process: requestSpec
				putResponseIn: responseSpec ]
		until: aBlock
! !

!MaRequestServerForTcp methodsFor: 'protocol' stamp: 'cmm 7/16/2002 21:41'!
isProtocolRequest: aByteArray
	"Any byteArray representation of a true request will have its first four occupied by the buffer size, so 'prot' should be a fairly safe, meaningful code.  That would indicate a nearly 2GB bufferSize..  I only hope our networks get fast enough..!!"

	| compareArray |
	compareArray _ #(112 114 111 116).  "character values for 'prot'."
	1
		to: 4
		do:
			[ :position | 
			(aByteArray at: position) == (compareArray at: position)
				ifFalse: [ ^false ] ].
	^true
! !

!MaRequestServerForTcp methodsFor: 'protocol' stamp: 'cmm 7/23/2002 19:43'!
protocolArray
	"If, for some reason, you wish to persist your request protocol (preserving their classId's), here is a ByteArray"

	^serializer classDefinitionsByteArray
! !

!MaRequestServerForTcp methodsFor: 'initializing' stamp: 'cmm 5/2/2002 23:50'!
protocol: aCollection
	"Establish the class ids simply by serializing aCollection."

	serializer serializeGraph: aCollection.
	serializer resetOidManager
! !

!MaRequestServerForTcp methodsFor: 'initializing' stamp: 'cmm 6/13/2002 00:33'!
requestByteArraySize: anInteger

	serializer _ MaObjectSerializer new: anInteger.
	socket _ MaSynchronousServerSocket
		requestSpecification:
			(MaByteArraySpecification new
				byteArray: serializer byteArray;
				yourself)
		responseSpecification: (MaByteArraySpecification new: serializer byteArray size)! !

!MaRequestServerForTcp methodsFor: 'initializing' stamp: 'cmm 4/7/2002 22:50'!
requestProcessor: someRequestProcessor
	"requestProcessor must respond to #protocol if it is to participate in as a *remote* requestProcessor server.  #protocol answers aCollection of all request classes and all response classes any client can expect to encounter in the full graph of the requests and responses."

	super requestProcessor: someRequestProcessor.
	self protocol: requestProcessor protocol
! !

!MaRequestServerForTcp methodsFor: 'private' stamp: 'cmm 6/12/2002 00:11'!
process: requestSpecification putResponseIn: responseSpecification
	"requestSpecification and responseSpecification are instances of MaByteArraySpecification"

	| saveByteArray requestByteArray response |
	(self isProtocolRequest: requestSpecification byteArray)
		ifTrue:
			[ responseSpecification
				byteArray: self protocolArray;
				useAllBytes.
			serializer resetOidManager.
			^self ].
	requestByteArray _ serializer materializeGraph: (serializer byteArray ~~ requestSpecification byteArray
		ifTrue:
			[ "A large request buffer caused us to have to grow.."
			saveByteArray _ serializer byteArray.
			requestSpecification byteArray ]
		ifFalse:
			[ serializer byteArray ]).
	response _ self processRequest: requestByteArray.
	serializer serializeGraph: response.
	responseSpecification
		byteArray: serializer byteArray;
		numberOfBytes: serializer graphBuffer bufferSize.		
	serializer
		byteArray: saveByteArray;
		resetOidManager
! !


!MaRequestServerForTcp class methodsFor: 'creation' stamp: 'cmm 6/9/2002 22:30'!
requestProcessor: requestProcessor

	^self
		requestProcessor: requestProcessor
		requestByteArraySize: 100000
! !

!MaRequestServerForTcp class methodsFor: 'creation' stamp: 'cmm 6/9/2002 22:30'!
requestProcessor: requestProcessor requestByteArraySize: anInteger
	"requestProcessor -> a class that implements processRequest: <one of your request objects>
	anInteger -> how much to preallocate for receiving serialized requests, as well as serializing the responses."

	^self new
		requestByteArraySize: anInteger;
		requestProcessor: requestProcessor
! !


!MaServerLink methodsFor: 'actions' stamp: 'cmm 4/8/2002 23:33'!
connect
	"Remote links have connection code they need to run.."
! !

!MaServerLink methodsFor: 'actions' stamp: 'cmm 4/1/2002 00:25'!
disconnect
	"Remote links need to clean up resources."
! !

!MaServerLink methodsFor: 'actions' stamp: 'cmm 3/30/2002 22:01'!
submit: aMaServerRequest

	self subclassResponsibility! !


!MaLocalServerLink methodsFor: 'actions' stamp: 'cmm 3/30/2002 22:03'!
submit: aMaServerRequest

	^localRequestServer processRequest: aMaServerRequest
! !

!MaLocalServerLink methodsFor: 'private' stamp: 'cmm 3/30/2002 22:52'!
localRequestServer: aMaLocalRequestServer

	localRequestServer _ aMaLocalRequestServer
! !


!MaLocalServerLink class methodsFor: 'as yet unclassified' stamp: 'cmm 3/30/2002 22:50'!
localRequestServer: aMaLocalRequestServer

	^self new localRequestServer: aMaLocalRequestServer
! !


!MaRemoteServerLink methodsFor: 'accessing' stamp: 'cmm 5/8/2002 23:24'!
hostAddress

	^socket hostAddress! !

!MaRemoteServerLink methodsFor: 'accessing' stamp: 'cmm 5/8/2002 23:24'!
port

	^socket port! !

!MaRemoteServerLink methodsFor: 'actions' stamp: 'cmm 7/16/2002 21:30'!
submit: aMaServerRequest

	^serializer
		resetOidManager;  "This is needed because we could potentially have oid conflicts from the previous materialization!!"
		materializeGraph: (self getByteArrayResponseFor: aMaServerRequest)
! !

!MaRemoteServerLink methodsFor: 'private' stamp: 'cmm 4/7/2002 23:44'!
bufferSize: anInteger

	self serializer: (MaObjectSerializer new: anInteger)
! !

!MaRemoteServerLink methodsFor: 'private' stamp: 'cmm 4/7/2002 23:43'!
getByteArrayResponseFor: aMaServerRequest

	serializer serializeGraph: aMaServerRequest.
	^socket
		sendData: serializer byteArray
		startingAt: 1
		count: serializer graphBuffer bufferSize
		waitForReplyIn: serializer byteArray 
		startingAt: 1
! !

!MaRemoteServerLink methodsFor: 'private' stamp: 'cmm 6/16/2002 21:32'!
getProtocolArray

	^socket
		sendData: 'prot' asByteArray
		startingAt: 1
		count: 4
		waitForReplyIn: serializer byteArray 
		startingAt: 1
! !

!MaRemoteServerLink methodsFor: 'private' stamp: 'cmm 3/30/2002 22:01'!
serializer: aMaObjectSerializer

	serializer := aMaObjectSerializer
! !

!MaRemoteServerLink methodsFor: 'private' stamp: 'cmm 6/7/2002 00:10'!
socket: aMaSynchronousSocket

	socket _ aMaSynchronousSocket! !

!MaRemoteServerLink methodsFor: 'connections' stamp: 'cmm 4/9/2002 00:28'!
connect

	serializer classDefinitionsByteArray: self getProtocolArray
! !

!MaRemoteServerLink methodsFor: 'connections' stamp: 'cmm 4/15/2002 00:18'!
disconnect

	socket disconnect
! !


!MaRemoteServerLink class methodsFor: 'creation' stamp: 'cmm 4/7/2002 23:42'!
hostAddress: aByteArray port: anInteger

	^self
		hostAddress: aByteArray
		port: anInteger
		bufferSize: 2000000
! !

!MaRemoteServerLink class methodsFor: 'creation' stamp: 'cmm 4/7/2002 23:41'!
hostAddress: aByteArray port: anInteger bufferSize: bufferSize
	"bufferSize is the size of the byteArray used to send requests and materialize responses."

	^self new
		bufferSize: bufferSize;
		socket:
			(MaSynchronousSocket
				hostAddress: aByteArray
				port: anInteger);
		yourself
! !


!MaServerRequest methodsFor: 'actions'!
process

	self subclassResponsibility
! !


!MaSynchronousSocket methodsFor: 'accessing' stamp: 'cmm 5/8/2002 23:23'!
hostAddress

	^hostAddress
! !

!MaSynchronousSocket methodsFor: 'accessing' stamp: 'cmm 5/8/2002 23:23'!
port

	^port! !

!MaSynchronousSocket methodsFor: 'send' stamp: 'cmm 6/17/2002 21:42'!
sendData: aStringOrByteArray startingAt: startIndex count: numberOfBytes waitForReplyIn: aMaByteArraySpecification
	"Use this method if you need to know how many bytes were received in the response."

	self
		connectTo: hostAddress
		port: port.
	^[ self
		sendData: aStringOrByteArray
		startingAt: startIndex
		count: numberOfBytes.
	self receiveInto: aMaByteArraySpecification ] 

		ensure: [ self closeAndDestroy ]
! !

!MaSynchronousSocket methodsFor: 'send' stamp: 'cmm 6/16/2002 21:28'!
sendData: aStringOrByteArray startingAt: startIndex count: numberOfBytes waitForReplyIn: answerByteArray startingAt: answerIndex
	"Perform a synchronous TCP call, sending aStringOrByteArray (from position startIndex) to my peer socket and then wait up to my timeout period for a reply, which I'll put in answerByteArray starting at answerIndex.  If I run out of room in answerByteArray, I'll create you a new one and return it, otherwise you'll get your own answerByteArray back.  If something goes wrong during all this, I'll signal an MaError."

	^self
		sendData: aStringOrByteArray
		startingAt: startIndex
		count: numberOfBytes
		waitForReplyIn: (MaByteArraySpecification new 
			byteArray: answerByteArray; 
			startPosition: startIndex;
			yourself)
! !

!MaSynchronousSocket methodsFor: 'server' stamp: 'cmm 3/22/2002 00:07'!
hostAddress: aByteArray

	hostAddress _ aByteArray! !

!MaSynchronousSocket methodsFor: 'server' stamp: 'cmm 3/22/2002 00:07'!
port: anInteger

	port _ anInteger! !

!MaSynchronousSocket methodsFor: 'private' stamp: 'cmm 6/10/2002 22:23'!
closeAndDestroy

	socket closeAndDestroy.
	socket _ nil  "Socket has a finalize method that releases system resources, make it garbage as soon as possible.."
! !

!MaSynchronousSocket methodsFor: 'private' stamp: 'cmm 7/23/2002 19:44'!
connectTo: addressByteArray port: portInteger

	hostAddress isNil | port isNil ifTrue: [ MaUserError signal: 'hostAddress and port are required.' ].
	socket _ Socket newTCP
		connectTo: addressByteArray
		port: portInteger.
	(socket waitForConnectionUntil: Socket standardDeadline)
		ifFalse:
			[ self error: 'server not responding' ]! !

!MaSynchronousSocket methodsFor: 'private' stamp: 'cmm 6/4/2002 00:02'!
disconnect

	socket close
! !

!MaSynchronousSocket methodsFor: 'private' stamp: 'cmm 7/23/2002 19:44'!
receiveInto: aMaByteArraySpecification
	"We are expecting a response from my server-peer.  Wait for data to be available and then milk it into aMaByteArraySpecification's byteArray starting at its indicated startPosition until there is nothing left to receive.  If its byteArray is not large enough, I will replace it with a bigger one."

	| currentPosition lastPacketSize |
	lastPacketSize _ -1.
	currentPosition _ aMaByteArraySpecification startPosition.
	[ lastPacketSize ~= 0 or: [ currentPosition = aMaByteArraySpecification startPosition ] ]
		whileTrue:
			[ lastPacketSize _ socket
				readInto: aMaByteArraySpecification byteArray
				startingAt: currentPosition.
			currentPosition _ currentPosition + lastPacketSize.
			(lastPacketSize > 0 and: [ currentPosition = (aMaByteArraySpecification byteArray size + 1) ])
				ifTrue:
					[ | newByteArray |
					aMaByteArraySpecification byteArray size > 10000000
						ifTrue: [ self maSignalSecurityViolation: 'Possible flood attack.' ].
					newByteArray _ ByteArray new: aMaByteArraySpecification byteArray size * 2.
					newByteArray
						replaceFrom: 1
						to: aMaByteArraySpecification byteArray size
						with: aMaByteArraySpecification byteArray
						startingAt: 1.
					aMaByteArraySpecification byteArray: newByteArray ] ].
	self maMarked: 'debug helper'.
	"self verify: currentByteArray."  "uncomment to verify.."
	aMaByteArraySpecification numberOfBytes: currentPosition - 1.
	^aMaByteArraySpecification byteArray! !

!MaSynchronousSocket methodsFor: 'private' stamp: 'cmm 7/23/2002 19:45'!
sendData: aStringOrByteArray startingAt: startIndex count: numberOfBytes
	"Send numberOfBytes of aStringOrByteArray starting at startIndex.
	Make multiple calls, if necessary, to send all of numberOfBytes."

	| bytesSent |
	bytesSent _ 0.
	[ bytesSent < numberOfBytes ]
		whileTrue:
			[ bytesSent _ bytesSent +
				(socket
					sendSomeData: aStringOrByteArray
					startIndex: bytesSent + 1
					count: numberOfBytes - bytesSent) ].
	socket waitForSendDoneUntil: Socket standardDeadline
! !

!MaSynchronousSocket methodsFor: 'private' stamp: 'cmm 4/12/2002 00:07'!
socket: aSocket

	socket _ aSocket! !

!MaSynchronousSocket methodsFor: 'debugging' stamp: 'cmm 7/23/2002 20:00'!
verify: aByteArray

	| fs fileBa |
	fs _ FileStream oldFileNamed: 'temp.dat'.
	fileBa _ fs next: fs size.
	fs close.
	(fileBa asByteArray = (aByteArray copyFrom: 1 to: fileBa size)) ifFalse: [ MaSoftwareError signal: 'byteArray''s don''t match..' ]! !

!MaSynchronousSocket methodsFor: 'debugging' stamp: 'cmm 4/18/2002 22:20'!
writeForVerification: aByteArray

	| fd fs |
	fd _ FileDirectory on: 's:\'.
	fd deleteFileNamed: 'temp.dat'.
	fs _ FileStream newFileNamed: 's:\temp.dat'.
	fs nextPutAll: aByteArray.
	fs close
! !

!MaSynchronousSocket methodsFor: 'printing' stamp: 'cmm 6/10/2002 22:45'!
printOn: aStream

	super printOn: aStream.
	aStream
		nextPutAll: ' (';
		nextPutAll: (socket ifNil: [ 'not connected' ] ifNotNil: [ socket statusString ]);
		nextPut: $)
! !


!MaSynchronousServerSocket methodsFor: 'private' stamp: 'cmm 6/13/2002 00:31'!
requestSpecification: aMaByteArrayRequest

	requestSpecification _ aMaByteArrayRequest! !

!MaSynchronousServerSocket methodsFor: 'private' stamp: 'cmm 6/13/2002 00:31'!
responseSpecification: aMaByteArrayRequest

	responseSpecification _ aMaByteArrayRequest! !

!MaSynchronousServerSocket methodsFor: 'initializing' stamp: 'cmm 6/13/2002 00:34'!
initialize

	super initialize.
	Socket initializeNetwork.
	socket _ Socket newTCP
! !

!MaSynchronousServerSocket methodsFor: 'initializing' stamp: 'cmm 7/24/2002 23:28'!
listenOn: portNumber do: twoArgBlock until: untilBlock
	"For every client connection, execute twoArgBlock with the receivers receiveBuffer with the contents sent from the client.  The first argument is the entire request byteArraySpec, an instance of MaByteArraySpecification, the second is another MaByteArraySpecification.  These each hold the byteArray, starting position and number of bytes.  These are useful for optimizing performance since it saves allocating a new ByteArray every time."

	| clientSocket savedRequestByteArray savedResponseByteArray |

	savedRequestByteArray _ requestSpecification byteArray.
	savedResponseByteArray _ responseSpecification byteArray.
	socket
		listenOn: portNumber
		backlogSize: 20.
	socket isValid ifFalse: [ self error: 'accept not supported..' ].
	[
	[ | done |
	done _ untilBlock value.
	clientSocket _ socket waitForAcceptUntil: Time millisecondClockValue + 1000.
	done ]
		whileFalse:
			[ (clientSocket notNil and: [ clientSocket isConnected ])
				ifTrue:
					[
						[ clientSocket _ clientSocket maAsSynchronousSocket.
						clientSocket receiveInto: requestSpecification.
						twoArgBlock
							value: requestSpecification
							value: responseSpecification.
						clientSocket
							sendData: responseSpecification byteArray
							startingAt: responseSpecification startPosition
							count: responseSpecification numberOfBytes ] 
								ensure:
									[ clientSocket closeAndDestroy.
									clientSocket _ nil.
									requestSpecification byteArray: savedRequestByteArray.
									responseSpecification byteArray: savedResponseByteArray ] ] ] ]
		ensure: [ socket closeAndDestroy ]
! !


!MaSynchronousSocket class methodsFor: 'private' stamp: 'cmm 4/12/2002 00:06'!
clientSocket: aSocket

	^self new socket: aSocket! !

!MaSynchronousSocket class methodsFor: 'creation' stamp: 'cmm 5/2/2002 23:38'!
hostAddress: aByteArray port: anInteger

	Socket initializeNetwork.
	^self new
		hostAddress: aByteArray;
		port: anInteger
! !


!MaSynchronousServerSocket class methodsFor: 'creation' stamp: 'cmm 6/13/2002 00:30'!
requestSpecification: requestSpec responseSpecification: responseSpec

	^self new
		requestSpecification: requestSpec;
		responseSpecification: responseSpec;
		yourself
! !


!Socket methodsFor: 'connection open/close' stamp: 'cmm 3/21/2002 23:48'!
accept
	"Accept a connection from the receiver socket.
	Return a new socket that is connected to the client"

	^self class acceptFrom: self.! !

!Socket methodsFor: 'sending-receiving' stamp: 'cmm 4/13/2002 00:07'!
readInto: aStringOrByteArray startingAt: aNumber
	"Read data into the given buffer starting at the given index and return the number of bytes received. Note the given buffer may be only partially filled by the received data."
self maMarked: 'dev'.  "Put the wait back to Socket standardDeadline"
	(self waitForDataUntil: Time millisecondClockValue +  172800000 "2 days") ifFalse: [self error: 'receive timeout'].
	^ self primSocket: socketHandle
		receiveDataInto: aStringOrByteArray
		startingAt: aNumber
		count: aStringOrByteArray size - aNumber + 1
! !

!Socket methodsFor: 'converting' stamp: 'cmm 4/12/2002 00:06'!
maAsSynchronousSocket

	^MaSynchronousSocket clientSocket: self
! !


!MaSynchronousServerSocket reorganize!
('start server')
('private' requestSpecification: responseSpecification:)
('initializing' initialize listenOn:do:until:)
!


!MaSynchronousSocket reorganize!
('accessing' hostAddress port)
('send' sendData:startingAt:count:waitForReplyIn: sendData:startingAt:count:waitForReplyIn:startingAt:)
('server' hostAddress: port:)
('private' closeAndDestroy connectTo:port: disconnect receiveInto: sendData:startingAt:count: socket:)
('initializing')
('debugging' verify: writeForVerification:)
('printing' printOn:)
!


!MaRemoteServerLink reorganize!
('accessing' hostAddress port)
('actions' submit:)
('private' bufferSize: getByteArrayResponseFor: getProtocolArray serializer: serverRefreshProtocol socket:)
('connections' connect disconnect)
!


!MaLocalServerLink class reorganize!
('as yet unclassified' localRequestServer:)
!


!MaLocalServerLink reorganize!
('actions' submit:)
('private' getRepositoryController localRequestServer:)
!


!MaServerLink reorganize!
('actions' connect disconnect submit:)
('private')
('private actions')
('initializing')
!


!MaRequestServerForTcp class reorganize!
('creation' requestProcessor: requestProcessor:requestByteArraySize:)
!


!MaLocalRequestServer class reorganize!
('creation')
!


!MaRequestServer reorganize!
('protocol')
('initializing' requestProcessor:)
('private' getRepositoryController processRequest:)
!


!MaByteArraySpecification reorganize!
('converting' logicalByteArray)
('accessing' byteArray byteArray: numberOfBytes numberOfBytes: startPosition startPosition: useAllBytes)
('initializing' initialize)
!

"Postscript:
Leave the line above, and replace the rest of this comment by a useful one.
Executable statements should follow this comment, and should
be separated by periods, with no exclamation points (!!).
Be sure to put any further comments in double-quotes, like this one."
!

