'From Squeak3.2gamma of 15 January 2002 [latest update: #4743] on 25 July 2002 at 11:25:26 pm'!
MaSingleton subclass: #MaDynamicVarContainer
	instanceVariableNames: 'vars '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma base classes-private'!

!MaDynamicVarContainer commentStamp: '<historical>' prior: 0!
The purpose of this is to store the instance-specific
(or "unique instance variables") of an object.  One
instance of MyClass can have a virtual instance variable
called #myVar, but another instance may not.

For virtual instance variables that affect every instance
of a class use MyClass>>#defineVirtualInstVarNamed:.

The implementation of the container is a weak-key identity
dictionary of IdentityDictionary's.

key							value
myObject					aDictionary
								key									value
								#var name						anObject
!

MaSingleton subclass: #MaVirtualInstanceVariableContainer
	instanceVariableNames: 'definitions variables '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma base classes-private'!

!MaVirtualInstanceVariableContainer commentStamp: 'cmm 7/22/2002 23:34' prior: 0!
Virtual Instance variables are variables that simulate
the presence of an instance variable.  It behaves like instance
variables in that you can't define two variables of the same name
in the same hierarchy.  You define them to a class and then all instances of that class
get the variable.

Virtual instance variables can be useful for adding instance variables to
vendor or base classes.  They can also be used to hide storage from object persistence frameworks, so you can have "transient" variables that won't be persisted.

This class is a Singleton that is used to contain all of the virtual
instance variables for all classes and instances in the system.  It also contains
the class definitions in terms of virtual instance variables.

Basically, it has two instance variables of its own, definitions and
variables.  Definitions defines the virtual variables for each class in
the system.

The format of it is:

key				value
aClass			Set(#var1 #var2 #var3)

The other is variables, it is a Dictionary of Dictionaries.

key				value
myObject		->  key							value		<------ (Another dictionary)
						#instVarName			valueObject

Note that garbage collection will operate normally only
when two objects don't virtually reference each other.
In other words, if thisObject virtually references thatObject,
and thatObject virtually references thisObject, neither one
will get garbage collected.
!


!Object methodsFor: 'accessing'!
maDynamicVarNamed: aString

	^self dynamicVarContainer
		varNamed: aString
		for: self
! !

!Object methodsFor: 'accessing'!
maVirtualInstVarNamed: aSymbol 
	"Return the value of the virtual instance variable named aSymbol.
	Virtual Instance variables are variables that are stored in a
	MaVirtualInstanceVariableContainer (a Singleton).
	They are used to simulate the adding
	of instance variables to classes which you normally cannot (or
	should not), such as Object."

	^self virtualInstVarContainer
		variableNamed: aSymbol
		for: self
! !

!Object methodsFor: 'building'!
maDeleteDynamicVarNamed: aString

	^self dynamicVarContainer removeVarNamed: aString
! !

!Object methodsFor: 'building'!
maDynamicVarNamed: aString put: anObject

	^self dynamicVarContainer
		dynamicVarNamed: aString
		put: anObject
		for: self
! !

!Object methodsFor: 'building'!
maVirtualInstVarNamed: aSymbol put: anObject 
	"Set the value of the virtual instance variable named aSymbol.
	Virtual Instance variables are variables that are stored in a global
	dictionary called VirtualInstanceVariables.  They are used to simulate
	the adding instance variables to Object (which is something you
	should not (and probably cannot) do)."

	self virtualInstVarContainer
		variableNamed: aSymbol
		put: anObject
		for: self
! !

!Object methodsFor: 'private accessing'!
dynamicVarContainer

	^MaDynamicVarContainer current
! !

!Object methodsFor: 'private accessing'!
virtualInstVarContainer

	^MaVirtualInstanceVariableContainer current
! !


!Behavior methodsFor: 'accessing'!
allVirtualInstVarNames

	^self virtualInstVarContainer variablesFor: self
! !

!Behavior methodsFor: 'testing'!
definesVirtualInstVarNamed: aSymbol 
	"Indicate whether the virtual variable named aSymbol is defined to the
	receiver."

	^self virtualInstVarContainer definesVariableNamed: aSymbol in: self! !

!Behavior methodsFor: 'testing'!
definesVirtualInstVars

	"Indicate whether this class defines any virtual instance variables."

	^self allVirtualInstVarNames notEmpty
! !

!Behavior methodsFor: 'actions'!
defineVirtualInstVarNamed: aSymbol

	"Define a virtual instance variable to the receiver named
	aSymbol.  This should only be used to 'add instance variables'
	to classes that ordinarily cannot (and even your reasons for
	doing that should be scrutinized).  Some frameworks may
	need something like this."

	self virtualInstVarContainer
		defineVariable: aSymbol
		to: self
! !

!Behavior methodsFor: 'actions'!
undefineAllVirtualInstVars
	"Remove all virtual instance variable definitions from the
	receiver class.  This will cause any information in those variables
	that may be present to be lost."

	self virtualInstVarContainer undefineAllVariablesIn: self
! !

!Behavior methodsFor: 'actions'!
undefineVirtualInstVarNamed: aSymbol

	"Undefine a virtual instance variable named aSymbol.
	Any information stored in instances of the receiver in
	the virtual instance variable will be lost."

	self virtualInstVarContainer
		undefineVariable: aSymbol
		in: self
! !


!MaDynamicVarContainer methodsFor: 'testing'!
definesVariableNamed: aString

	^vars includesKey: aString
! !

!MaDynamicVarContainer methodsFor: 'accessing'!
dynamicVarNamed: aSymbol for: holderObject

	^(vars at: holderObject) at: aSymbol
! !

!MaDynamicVarContainer methodsFor: 'building'!
dynamicVarNamed: aSymbol put: anObject for: holderObject

	(vars
		at: holderObject
		ifAbsentPut: [ Dictionary new ])
			at: aSymbol
			put: anObject
! !

!MaDynamicVarContainer methodsFor: 'initializing'!
initialize

	vars := WeakIdentityKeyDictionary new
! !


!MaVirtualInstanceVariableContainer methodsFor: 'testing'!
definesVariableNamed: aSymbol in: aClass 

	"Return true if variable named aSymbol as been defined to aClass,
	or any of its superclasses.  See method comment for details."

	aClass withAllSuperclasses
		do: [ :each | (self
			explicitlyDefinesVariableNamed: aSymbol
			inClass: each) ifTrue: [ ^true ] ].
	^false
! !

!MaVirtualInstanceVariableContainer methodsFor: 'private testing' stamp: 'cmm 2/27/2002 20:42'!
definesVariableNamed: aSymbol  inHierarchyContainingClass: aClass

	"Answer true if any class in the same inheritance hierarchy as aClass
	defines aSymbol as a virtual instance variable."

	(aClass allSubclasses addAll: aClass allSuperclasses; yourself) do:
		[ :each | (self
			explicitlyDefinesVariableNamed: aSymbol
			inClass: each) ifTrue: [ ^true ] ].
	^false
! !

!MaVirtualInstanceVariableContainer methodsFor: 'private testing'!
explicitlyDefinesVariableNamed: aString inClass: aClass

	"Answer true if aClass is the exact class that defines aString"

	^(definitions includesKey: aClass)
		and: [ (definitions at: aClass) includes: aString ]
! !

!MaVirtualInstanceVariableContainer methodsFor: 'building'!
defineVariable: aSymbol to: aClass

	"Add virtual storage registered under the name aSymbol
	to aClass."

	(self
		definesVariableNamed: aSymbol 
		inHierarchyContainingClass: aClass) ifTrue:
			[ self error: 'Cannot define ', aSymbol, ' to ', aClass name asString, ' because it is already defined in the hierarchy containing ', aClass name ].
	(definitions
		at: aClass
		ifAbsentPut: [ IdentitySet new ]) add: aSymbol
! !

!MaVirtualInstanceVariableContainer methodsFor: 'private initializing' stamp: 'cmm 4/20/2002 00:52'!
initialize

	super initialize.
	definitions := IdentityDictionary new.
	variables := WeakIdentityKeyDictionary new
! !

!MaVirtualInstanceVariableContainer methodsFor: 'private actions' stamp: 'cmm 7/22/2002 23:36'!
install: anObject

	"Adds anObject to the (outer) Dictionary."

	^variables
		at: anObject
		put: Dictionary new
! !

!MaVirtualInstanceVariableContainer methodsFor: 'actions'!
undefineAllVariablesIn: aClass
	"Remove virtual storage registered under the name aSymbol
	to aClass.  All objects which store this variable will no longer
	be able to access it."

	definitions
		removeKey: aClass
		ifAbsent: [ "do nothing" ]
! !

!MaVirtualInstanceVariableContainer methodsFor: 'actions'!
undefineVariable: aSymbol in: aClass

	"Remove virtual storage registered under the name aSymbol
	to aClass.  All objects which store this variable will no longer
	be able to access it."

	self
		undefineVariable: aSymbol
		in: aClass
		ifAbsent: [ self error: aSymbol asString, ' was not defined to ', aClass name printString ]
! !

!MaVirtualInstanceVariableContainer methodsFor: 'accessing'!
variableNamed: aSymbol for: anObject

	"Return the value of anObjects instance variable named aString.
	This looks a little
	strange at first (the nested return), but it does work, and we need
	maximum performance."

	(self
		definesVariableNamed: aSymbol
		in: anObject class) ifFalse: 
			[ anObject
				error: anObject class name , ' does not virtually understand ' , aSymbol ].
	^(variables
		at: anObject
		ifAbsent: [^nil])
			at: aSymbol
			ifAbsent: [ nil ]
! !

!MaVirtualInstanceVariableContainer methodsFor: 'accessing'!
variableNamed: aSymbol put: aValue for: anObject
	"Set the value of anObjects instance variable named aSymbol
	to aValue."

	(self
		definesVariableNamed: aSymbol
		in: anObject class)
			ifFalse:
				[ anObject error: anObject class name , ' does not understand ' , aSymbol asString ].
	(variables
		at: anObject
		ifAbsent: [ self install: anObject ])
			at: aSymbol
			put: aValue
! !

!MaVirtualInstanceVariableContainer methodsFor: 'accessing'!
variablesFor: aClass 

	"Return the set of virtual instance variables for aClass.
	This will probably not be needed for general usage."

	|vars|

	vars := Set new.
	aClass withAllSuperclasses do: [ :each | (definitions includesKey: each)
		ifTrue: [ vars addAll: (definitions at: each) ] ].
	^vars
! !


!MaVirtualInstanceVariableContainer class methodsFor: 'framework'!
reset

	current := self new
! !

"Postscript:
"
MaVirtualInstanceVariableContainer reset.!

