'From Squeak3.2gamma of 15 January 2002 [latest update: #4743] on 25 July 2002 at 11:25:23 pm'!
MaObject subclass: #MaIntervalCollection
	instanceVariableNames: 'intervals proximityThreshold '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma base classes-Special collections'!

!MaIntervalCollection commentStamp: 'cmm 7/22/2002 23:39' prior: 0!
This class keeps track of a collection of intervals.

As intervals are added that intersect or are within the proximityThreshold
of existing intervals, the existing interval is enlarged to include for the new interval.

If an interval is removed in the middle of a large interval such that
a "hole" is created, a new interval will be created and the other one
will be adjusted so that they will represent only the valid ranges.

For example:

	if an instance of this class has 1..10

	and we add: 8..15, then this class will have a single interval from 1..15

	Then remove 5..11..  Assuming a proximityThreshold of 1, this class will have the following two intervals:

		1..4
		12..15

Use a proximityThreshold of 0 to track the real-numbers..!

MaObject subclass: #MaPermutator
	instanceVariableNames: 'arguments '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma base classes-Special collections'!
MaObject variableSubclass: #MaRoundCollection
	instanceVariableNames: 'array head tail '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma base classes-Special collections'!

!MaRoundCollection commentStamp: '<historical>' prior: 0!
(self new: 5) addAll: ($a to: $f)

yields:

#($b $c $d $e $f)

but physically, it is:

#($f $b $c $d $e)
!

MaObject subclass: #MaTree
	instanceVariableNames: 'parents '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Ma base classes-Special collections'!

!MaIntervalCollection methodsFor: 'building'!
add: anInterval

	self
		addFrom: anInterval first
		to: anInterval last! !

!MaIntervalCollection methodsFor: 'building'!
addFrom: from to: to

	| physicalFromIndex physicalToIndex |

	physicalFromIndex := self physicalIndexOf: from.
	physicalToIndex := self physicalIndexOf: to.
	(from = to
		and: [ self isStartOfGap: physicalFromIndex ])
			ifTrue:
				[ "User does not realize that this class deals with real-number-based
				intervals that are inclusive on the left and exclusive on the right.
				Take care of him by adding a sensible interval."
				^self addFrom: from to: to + proximityThreshold ].
	physicalFromIndex = physicalToIndex
		ifTrue:
			[ "if in a gap, add a new interval"
			((self isStartOfGap: physicalFromIndex)
				and: [ (self
					fromValue: from
					isOutsideProximityThresholdAt: physicalFromIndex)
				and:
					[ self
						toValue: to
						isOutsideProximityThresholdAt: physicalToIndex ] ])
				ifTrue:
					[ self
						privateInsertFrom: from
						to: to
						afterIndex: physicalFromIndex.
					^from ] ].
	self
		extendIntervalFrom: from
		at: physicalFromIndex
		to: to
		at: physicalToIndex.
	^from! !

!MaIntervalCollection methodsFor: 'building'!
remove: anInterval

	^self
		removeFrom: anInterval first
		to: anInterval last! !

!MaIntervalCollection methodsFor: 'building'!
removeFrom: from to: to

	| physicalFromIndex physicalToIndex |

	from = to
		ifTrue:
			[ "User does not realize that this class deals with real-number-based
			intervals that are inclusive on the left and exclusive on the right.
			Take care of him by removing a sensible interval."
			^self
				removeFrom: from
				to: to + proximityThreshold ].
	physicalFromIndex := self physicalIndexOf: from.
	physicalToIndex := self physicalIndexOf: to.
	physicalFromIndex = physicalToIndex
		ifTrue:
			[ "If completely inside an interval, we must add a new interval to break it up"
			((self isStartOfInterval: physicalFromIndex)
				and: [ (intervals at: physicalFromIndex + 1) > to ])
					ifTrue:
						[ self
							privateInsertFrom: from
							to: to
							afterIndex: physicalFromIndex.
						^from ] ].
	self
		extendGapFromValue: from
		at: physicalFromIndex
		toValue: to
		at: physicalToIndex.
	^from! !

!MaIntervalCollection methodsFor: 'private building' stamp: 'cmm 2/20/2002 01:17'!
bridgeFrom: physicalBridgeFromIndex to: physicalBridgeToIndex
	"Remove intervals in the receiver that are between physicalFromIndex
	and physicalToIndex which should be merged into one larger interval.
	The caller must ensure the from is the start of an interval and the to is
	the start of a gap."

	physicalBridgeToIndex - physicalBridgeFromIndex - 1
		timesRepeat: [ intervals removeAt: physicalBridgeFromIndex + 1 ]! !

!MaIntervalCollection methodsFor: 'private building'!
extendGapFromValue: from at: physicalFromIndex toValue: to at: physicalToIndex

	| voidStart voidEnd |

	voidStart := (self isStartOfInterval: physicalFromIndex)
		ifTrue:
			[ "Extend to of this gap back a little."
			intervals
				at: physicalFromIndex + 1
				put: from.
			physicalFromIndex + 1 ]
		ifFalse: [ physicalFromIndex ].
	voidEnd := ((self isStartOfInterval: physicalToIndex)
		and: [ physicalToIndex > physicalFromIndex  "account for when to is an endpoint exactly on the end of an interval" ])
			ifTrue:
				[ "Account for removing exact endpoints of an existing interval."
				(intervals at: physicalToIndex + 1) = to
					ifTrue: [ physicalToIndex + 2 ]
					ifFalse:
						[ "Extend from of next gap up a little.."
						intervals
							at: physicalToIndex
							put: to.
						physicalToIndex ] ]
			ifFalse: [ physicalToIndex + 1 ].
	self
		voidFrom: voidStart
		to: voidEnd! !

!MaIntervalCollection methodsFor: 'private building'!
extendIntervalFrom: from at: physicalFromIndex to: to at: physicalToIndex
	
	| bridgeStart bridgeEnd |

	bridgeStart := (self isStartOfGap: physicalFromIndex)
		ifTrue:
			[ "Extend from of next interval back a little.."
			(physicalFromIndex < self physicalSize and: [ (intervals at: physicalFromIndex + 1) <= (to + proximityThreshold) ])
				ifTrue:
					[ intervals
						at: physicalFromIndex + 1 "start of next interval"
						put: from ].
			physicalFromIndex > 0
				ifTrue: [ (from - proximityThreshold) <= (intervals at: physicalFromIndex)
					ifTrue: [ physicalFromIndex - 1 ]
					ifFalse: [ physicalFromIndex + 1 ] ]
				ifFalse: [ 1 ] ]
		ifFalse:
			[ bridgeStart := physicalFromIndex ].
	bridgeEnd := (self isStartOfGap: physicalToIndex)
		ifTrue:
			[ "Extend to of this interval forward a little.."
			(physicalToIndex > 0 and: [ (intervals at: physicalToIndex) >= (from - proximityThreshold) ])
				ifTrue:
					[ intervals
						at: physicalToIndex  "end of this interval"
						put: to ].
			physicalToIndex < self physicalSize
				ifTrue: [ (to + proximityThreshold) >= (intervals at: physicalToIndex + 1)
					ifTrue: [ physicalToIndex + 2 ]
					ifFalse: [ physicalToIndex ] ]
				ifFalse: [ self physicalSize ] ]
		ifFalse:
			[ bridgeEnd := physicalToIndex + 1 ].
	self
		bridgeFrom: bridgeStart
		to: bridgeEnd! !

!MaIntervalCollection methodsFor: 'private building'!
privateInsertFrom: low to: high afterIndex: physicalIndex

	intervals
		add: high
		afterIndex: physicalIndex.
	intervals
		add: low
		afterIndex: physicalIndex! !

!MaIntervalCollection methodsFor: 'private building'!
proximityThreshold: aMagnitude

	proximityThreshold := aMagnitude! !

!MaIntervalCollection methodsFor: 'private building' stamp: 'cmm 7/3/2002 00:13'!
voidFrom: voidStart to: voidEnd

	voidEnd - voidStart - 1
		timesRepeat: [ intervals removeAt: voidStart + 1 ]! !

!MaIntervalCollection methodsFor: 'private testing'!
fromValue: aMagnitude isOutsideProximityThresholdAt: physicalFromIndex

	^physicalFromIndex = 0
		or: [ (aMagnitude - proximityThreshold) > (intervals at: physicalFromIndex) ]! !

!MaIntervalCollection methodsFor: 'private testing'!
isStartOfGap: physicalIndex

	^physicalIndex even! !

!MaIntervalCollection methodsFor: 'private testing'!
isStartOfInterval: physicalIndex

	^physicalIndex odd! !

!MaIntervalCollection methodsFor: 'private testing'!
toValue: aMagnitude isOutsideProximityThresholdAt: physicalToIndex

	^physicalToIndex = self physicalSize
		or: [ (aMagnitude + proximityThreshold) < (intervals at: physicalToIndex + 1) ]! !

!MaIntervalCollection methodsFor: 'accessing'!
gapsDetect: twoArgBlock ifFound: foundBlock ifNone: noneBlock

	self gapsDo:
		[ :eachLow :eachHigh |
		(twoArgBlock
			value: eachLow
			value: eachHigh)
				ifTrue:
					[ ^foundBlock
						value: eachLow
						value: eachHigh ] ].
	^noneBlock value! !

!MaIntervalCollection methodsFor: 'accessing'!
highest

	^intervals isEmpty ifTrue: [ self error: self printString, ' is empty.' ] ifFalse: [ intervals last ]! !

!MaIntervalCollection methodsFor: 'accessing'!
indexOf: aMagnitude
	"Answer the index of the interval which contains aMagnitude.
	If aMagnitude is not part of any interval in the receiver, answer 0."

	| index |

	index := self physicalIndexOf: aMagnitude.
	^((self isStartOfInterval: index)
		and: [ (intervals at: index + 1) > aMagnitude ])
			ifTrue:
				[ self logicalIndexFor: index ]
			ifFalse:
				[ 0 ]! !

!MaIntervalCollection methodsFor: 'accessing'!
maDetect: twoArgBlock ifFound: foundBlock ifNone: noneBlock 
	self intervalsDo: 
			[:eachLow :eachHigh | 
			(twoArgBlock value: eachLow value: eachHigh) 
				ifTrue: [^foundBlock value: eachLow value: eachHigh]].
	^noneBlock value! !

!MaIntervalCollection methodsFor: 'accessing'!
size

	^intervals size / 2! !

!MaIntervalCollection methodsFor: 'actions'!
gapsDo: twoArgBlock
	"For each gap in the receivers intervals, value twoArgBlock.
	The first argument is the low of the gap, the second is the high."

	2
		to: intervals size - 2
		by: 2
		do:
			[ :eachLowIndex |
			twoArgBlock
				value: (intervals at: eachLowIndex)
				value: (intervals at: eachLowIndex + 1) ]! !

!MaIntervalCollection methodsFor: 'actions'!
intervalsDo: twoArgBlock
	"For each interval in the receivers intervals, value twoArgBlock.
	The first argument is the low of the interval, the second is the high."

	1
		to: intervals size - 1
		by: 2
		do:
			[ :eachLowIndex |
			twoArgBlock
				value: (intervals at: eachLowIndex)
				value: (intervals at: eachLowIndex + 1) ]! !

!MaIntervalCollection methodsFor: 'testing'!
includes: aMagnitude

	^(self indexOf: aMagnitude) > 0! !

!MaIntervalCollection methodsFor: 'testing'!
isEmpty

	^intervals isEmpty! !

!MaIntervalCollection methodsFor: 'testing'!
notEmpty

	^self isEmpty not
! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 4/23/2002 19:37'!
independentCopy

	^self copyWithAttributes: #('intervals')
! !

!MaIntervalCollection methodsFor: 'private' stamp: 'cmm 7/17/2002 22:53'!
initialize

	super initialize.

	"The intervals is stored in sorted order.  Each odd is the low of an interval, the
	even is the high."
	intervals := OrderedCollection new: 2.

	"proximityThreshold is a value used to determine how close the
	To end of an interval needs to be to the From end of the next
	interval before they will be merged into a single interval.  The
	default is 0 (no merging)."
	proximityThreshold := 0! !

!MaIntervalCollection methodsFor: 'private'!
logicalIndexFor: physicalIndex

	^(physicalIndex + 1) // 2! !

!MaIntervalCollection methodsFor: 'private'!
physicalIndexFor: logicalIndex

	^(logicalIndex * 2 - 1) max: 0! !

!MaIntervalCollection methodsFor: 'private'!
physicalIndexOf: aMagnitude

	| index |

	index := intervals size.
	[ index > 0 and: [ (intervals at: index) >= aMagnitude ] ] whileTrue: [ index := index - 1 ].
	^index! !

!MaIntervalCollection methodsFor: 'private'!
physicalSize

	^intervals size! !

!MaIntervalCollection methodsFor: 'printing' stamp: 'cmm 4/22/2002 22:25'!
maPrintAbbreviatedOn: aStream

	1
		to: intervals size
		do:
			[ :index |
			aStream nextPutAll: (intervals at: index) printString.
			index odd ifTrue: [ aStream nextPutAll: '..' ] ifFalse: [ aStream nextPutAll: '  ' ] ]
! !


!MaIntervalCollection class methodsFor: 'creation'!
proximityThreshold: aMagnitude

	^self new proximityThreshold: aMagnitude! !


!MaPermutator methodsFor: 'building'!
addArguments: aSequenceableCollection

	^self arguments add: aSequenceableCollection
! !

!MaPermutator methodsFor: 'building'!
arguments: aSequenceableCollectionOfSequenceableCollections

	arguments := aSequenceableCollectionOfSequenceableCollections
! !

!MaPermutator methodsFor: 'accessing'!
arguments

	^arguments! !

!MaPermutator methodsFor: 'accessing'!
size

	^self arguments
		inject: 1
		into: [ :prod :each | each size * prod ]
! !

!MaPermutator methodsFor: 'private initializing'!
initialize

	arguments := OrderedCollection new
! !

!MaPermutator methodsFor: 'actions'!
permutationsDo: oneArgBlock
	"For each combination of elements in the receivers args, value aBlock
	with the arguments of each combination of elements.  The argument to
	oneArgBlock is the Array of parameters for each combination."

	| args |

	args := Array new: self arguments size.
	1
		to: self size
		do:
			[ :elementIndex |
			1
				to: self arguments size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := self arguments at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (self arguments at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			oneArgBlock value: args ]
! !

!MaPermutator methodsFor: 'actions'!
permutationsWhere: conditionBlock do: oneArgBlock
	"Same as permutationsDo: except the conditionBlock must also be
	satisfied for aBlock to run.  This is only for mildly better looking code
	since a case statement could achieve the same effect."

	| args |

	args := Array new: self arguments size.
	1
		to: self size
		do:
			[ :elementIndex |
			1
				to: self arguments size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := self arguments at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (self arguments at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			(conditionBlock value: args) ifTrue: [ oneArgBlock value: args ] ]
! !


!MaPermutator class methodsFor: 'actions'!
permutationsOf: anArrayOfArrays do: aBlock
	"For each combination of elements in anArrayOfArray's, value aBlock
	with the arguments of each combination of elements.  There must be
	as many arguments in aBlock as there are elements in the outer array
	of Array's.

	Note, this also works with any sequenceable collection of sequenceable
	collections."

	| numberOfCombinations args |

	numberOfCombinations :=
		anArrayOfArrays
			inject: 1
			into: [ :prod :each | each size * prod ].
	args := Array new: anArrayOfArrays size.
	1
		to: numberOfCombinations
		do:
			[ :elementIndex |
			1
				to: anArrayOfArrays size
				do:
					[ :innerIndex | | eachInnerArray radix |
					eachInnerArray := anArrayOfArrays at: innerIndex.
					radix := 1.
					1
						to: innerIndex - 1
						do: [ :eachOrder | radix := radix * (anArrayOfArrays at: eachOrder) size ].
					args
						at: innerIndex
						put: (eachInnerArray at: (elementIndex // radix \\ eachInnerArray size + 1)) ].
			aBlock valueWithArguments: args ]
! !


!MaRoundCollection methodsFor: 'building'!
add: anObject 

	tail := tail + 1.
	^array at: self logicalTail put: anObject
! !

!MaRoundCollection methodsFor: 'building'!
at: anInteger put: anObject

	^array
		at: (self logicalIndexOf: anInteger)
		put: anObject
! !

!MaRoundCollection methodsFor: 'building'!
removeFirst

	| removed index |

	removed := array at: (index := self logicalHead).
	"Allow object to become garbage, like any other collection would."
	array
		at: index
		put: nil.
	head := head + 1.
	^removed
! !

!MaRoundCollection methodsFor: 'building'!
removeLast

	| removed logicalIndex |

	tail = 0 ifTrue: [ self error: self printString , ' is empty, so removeFirst is inappropriate.' ].
	removed := array at: (logicalIndex := self logicalTail).
	"Allow object being removed to become garbage."
	array
		at: logicalIndex
		put: nil.
	tail := tail - 1.
	^removed
! !

!MaRoundCollection methodsFor: 'accessing'!
at: anInteger

	^array at: (self logicalIndexOf: anInteger)
! !

!MaRoundCollection methodsFor: 'accessing'!
first

	^array at: self logicalHead
! !

!MaRoundCollection methodsFor: 'accessing'!
historicalSize
	"Every time add: is called, historicalSize will increase by 1.
	Every time removeLast is called, historicalSize will decrease by 1.
	Any other mutating operation will not affect historicalSize."

	^tail
! !

!MaRoundCollection methodsFor: 'accessing'!
isEmpty

	^self size = 0
! !

!MaRoundCollection methodsFor: 'accessing'!
last

	^array at: self logicalTail
! !

!MaRoundCollection methodsFor: 'accessing'!
maxSize
	^array size! !

!MaRoundCollection methodsFor: 'accessing'!
size

	^tail - head + 1
! !

!MaRoundCollection methodsFor: 'actions'!
do: aBlock

	head
		to: tail
		do: [ :index | aBlock value: (array at: (self logicalIndexOf: index)) ]
! !

!MaRoundCollection methodsFor: 'testing'!
includes: anObject

	self do: [ :each | each = anObject ifTrue: [ ^true ] ].
	^false
! !

!MaRoundCollection methodsFor: 'private initializing'!
initialize

	head := 1.
	tail := 0
! !

!MaRoundCollection methodsFor: 'private accessing'!
logicalHead

	| size logicalHead |

	^(logicalHead := head \\ (size := self maxSize)) = 0
		ifTrue: [ size ]
		ifFalse: [ logicalHead ]
! !

!MaRoundCollection methodsFor: 'private accessing'!
logicalIndexOf: anInteger

	| logicalIndex size |

	^(logicalIndex := head + anInteger - 1 \\ (size := self maxSize)) = 0
		ifTrue: [ size ]
		ifFalse: [ logicalIndex ]
! !

!MaRoundCollection methodsFor: 'private accessing'!
logicalTail

	| size logicalTail |

	^(logicalTail := tail \\ (size := self maxSize)) = 0
		ifTrue: [ size ]
		ifFalse: [ logicalTail ]
! !

!MaRoundCollection methodsFor: 'printing'!
printOn: aStream
	"Print up to 1000 elements."

	super printOn: aStream.
	[ aStream nextPutAll: ' ('.
	1
		to: (self size min: 100)
		do:
			[ :index |
			(self at: index) printOn: aStream.
			aStream nextPut: $  ].
	self size > 100 ifTrue: [ aStream nextPutAll: '... etc ...' ].
	aStream nextPut: $) ] critical
! !

!MaRoundCollection methodsFor: 'private building'!
size: anInteger

	array := Array new: anInteger
! !


!MaRoundCollection class methodsFor: 'creation'!
new

	^self shouldNotImplement
! !

!MaRoundCollection class methodsFor: 'creation'!
new: anInteger

	^super new size: anInteger
! !


!MaTree methodsFor: 'building'!
add: anObject
	"Adds anObject as a main parent in the receiver."

	self
		add: anObject
		asChildOf: nil.
	^anObject
! !

!MaTree methodsFor: 'building'!
add: anObject asChildOf: aParentObject 
	"Adds anObject as a child of aParentObject and then returns anObject."

	^parents at: anObject put: aParentObject! !

!MaTree methodsFor: 'building'!
addAll: aCollection

	"Adds all in aCollection as a main parents in the receiver."

	aCollection do: [ :each | self add: each ]
! !

!MaTree methodsFor: 'building'!
addPath: anOrderedCollection

	"Adds each object in anOrderedCollection as a new path in
	the receiver.  The first element in anOrderedCollection is the
	main parent, and each subsequent element is a child of the
	previous.  If part of the path already exists, there it is
	replaced (though they should be equivalent anyway)."

	^self
		addPath: anOrderedCollection
		to: nil
! !

!MaTree methodsFor: 'building'!
addPath: anOrderedCollection to: parentObject
	"Adds each object in anOrderedCollection as a new path in
	the receiver.  The first element in anOrderedCollection is the
	main parent, and each subsequent element is a child of the
	previous.  If part of the path already exists, there it is
	replaced (though they should be equivalent anyway)."

	anOrderedCollection
		inject: parentObject
		into:
			[ :parent :each |
			self
				add: each
				asChildOf: parent.
			each ].
	^anOrderedCollection
! !

!MaTree methodsFor: 'building'!
addTree: aMaTree

	^self addTree: aMaTree to: nil
! !

!MaTree methodsFor: 'building'!
addTree: aMaTree to: anObject

	(self includes: anObject) ifFalse: [ self error: anObject printString, ' is not part of ', self printString ].
	aMaTree doWithParents: [ :each :eachParent | self add: each asChildOf: eachParent ].
	^aMaTree
! !

!MaTree methodsFor: 'building' stamp: 'cmm 4/23/2002 19:40'!
copy

	^self copyWithAttributes: #('parents')
! !

!MaTree methodsFor: 'building'!
move: anObject toParent: newParentObject 
	^parents at: anObject put: newParentObject! !

!MaTree methodsFor: 'building' stamp: 'cmm 7/25/2002 19:42'!
remove: anObject
	"Removes anObject and all of its children."

	^self
		remove: anObject
		ifAbsent: [ self error: anObject printString, ' was not found in ', self printString ]
! !

!MaTree methodsFor: 'building'!
remove: anObject ifAbsent: oneArgBlock 
	"Answers the collection of removed objects removed after removing
	anObject after removing it and all of its children."

	| tree |
	(tree := self treeAt: anObject) allMembers 
		do: [:each | parents removeKey: each]! !

!MaTree methodsFor: 'enumerating'!
allLeafsDo: aBlock
	"Value aBlock with each element in the receiver that has no children."

	| children |

	children := self children.
	parents keysAndValuesDo:
		[ :eachChild :eachParent |
		(children includesKey: eachChild) ifFalse: [ aBlock value: eachChild ] ]
! !

!MaTree methodsFor: 'enumerating'!
depthFirstDo: twoArgBlock
	"Value twoArgBlock with each object in the receiver and its parent in the order
	of depthFirst."

	self roots do:
		[ :each |
		twoArgBlock
			value: each
			value: nil.
		self
			depthFirstDo: twoArgBlock
			at: each ]
! !

!MaTree methodsFor: 'enumerating'!
depthFirstDo: twoArgBlock at: anObject

	(self childrenOf: anObject) do:
		[ :each |
		twoArgBlock
			value: each
			value: anObject.
		self
			depthFirstDo: twoArgBlock
			at: each ]
! !

!MaTree methodsFor: 'enumerating'!
do: oneArgBlock

	self keysDo: oneArgBlock
! !

!MaTree methodsFor: 'enumerating'!
maDetect: aBlock ifFound: foundBlock ifNone: exceptionBlock 
	^self allMembers 
		maDetect: aBlock
		ifFound: foundBlock
		ifNone: exceptionBlock value! !

!MaTree methodsFor: 'enumerating'!
nodesInPathFrom: someChildObject to: someParentObject do: oneArgBlock
	"For each node in the tree from someChildObject to someParentObject, value
	oneArgBlock.  The highest level parent (nil) will not be used as an argument
	to value oneArgBlock."

	| node |

	node := someChildObject.
	oneArgBlock value: someChildObject.
	[ node isNil or: [ node = someParentObject ] ] whileFalse:
		[ oneArgBlock value: node.
		node := self parentOf: node ].
	someParentObject notNil ifTrue: [ oneArgBlock value: someParentObject ]
! !

!MaTree methodsFor: 'enumerating'!
pathFrom: anObject do: aBlock
	"Value aBlock with the objects from anObject to its highest level parent."

	^self
		nodesInPathFrom: anObject
		to: nil
		do: aBlock
! !

!MaTree methodsFor: 'accessing'!
allLeafs

	| answer |

	answer := Set new.
	self allLeafsDo: [ :each | answer add: each ].
	^answer
! !

!MaTree methodsFor: 'accessing'!
allMembers
	"Returns a Set of all the members of the hierarchy"

	^parents keys! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 1/23/2002 14:30'!
children

	| children |

	children := Dictionary new: parents size.
	parents keysAndValuesDo:
		[ :eachChild :eachParent |
		(children
			at: eachParent
			ifAbsentPut: [ OrderedCollection new ])
				add: eachChild ].
	^children
! !

!MaTree methodsFor: 'accessing'!
childrenOf: anObject 
	^parents maAllKeysAtValue: anObject! !

!MaTree methodsFor: 'accessing'!
collect: aBlock
	"Returns a new tree whose nodes have all been operated on by aBlock."

	^self
		collect: aBlock
		into: self class new
! !

!MaTree methodsFor: 'accessing'!
conform: aBlock 
	parents keysDo: [:each | (aBlock value: each) ifFalse: [^false]].
	^true! !

!MaTree methodsFor: 'accessing'!
depthOf: anObject
	"Returns the number of parents anObject has."

	| depth prevChild |

	(self includes: anObject)
		ifFalse:
			[ self error: self printString , ' does not include ' , anObject printString ].
	depth := 0.
	prevChild := anObject.
	[ prevChild isNil ] whileFalse:
		[ prevChild := self parentOf: prevChild.
		depth := depth + 1 ].
	^depth
! !

!MaTree methodsFor: 'accessing'!
detect: aBlock

	^self allMembers detect: aBlock
! !

!MaTree methodsFor: 'accessing'!
detect: aBlock ifNone: exceptionBlock

	^self allMembers
		detect: aBlock
		ifNone: exceptionBlock value
! !

!MaTree methodsFor: 'accessing'!
nodesInPathFrom: someChildObject to: someParentObject
	"Returns a collection of nodes between someChildObject
	and someParentObject."

	|answer node|

	answer := OrderedCollection new.
	self
		nodesInPathFrom: someChildObject
		to: someParentObject
		do: [ :each | answer add: each ].
	^answer
! !

!MaTree methodsFor: 'accessing'!
parentOf: anObject 
	^parents at: anObject! !

!MaTree methodsFor: 'accessing'!
pathFrom: anObject
	"Returns a collection of the objects that are contained in
	the path to anObject ordered from an Object to the top of the hierarchy."

	^self
		nodesInPathFrom: anObject
		to: nil
! !

!MaTree methodsFor: 'accessing'!
pathTo: anObject
	"Returns a collection of the objects that are contained in
	the path to anObject, ordered from the the top
	of the hierarchy to anObject."

	^(self pathFrom: anObject) reverse
! !

!MaTree methodsFor: 'accessing'!
reject: aBlock
	"Returns a new Tree with all branches removed that
	meet aBlock criteria.  See method comment for more
	info."

	| members newTree |

	newTree := self class new.
	(self allMembers reject: aBlock) do: [ :each | newTree addPath: (self pathTo: each) ].
	^newTree
! !

!MaTree methodsFor: 'accessing'!
roots
	"Answers a collection of all objects that have no parents
	(e.g. all MAIN parents)"

	^self childrenOf: nil
! !

!MaTree methodsFor: 'accessing'!
select: aBlock
	"Returns a new Tree with all branches removed that
	meet aBlock criteria.  See method comment for more
	info."

	| members newTree |

	newTree := self class new.
	(self allMembers select: aBlock) do: [ :each | newTree addPath: (self pathTo: each) ].
	^newTree
! !

!MaTree methodsFor: 'accessing'!
size
	^parents size! !

!MaTree methodsFor: 'accessing' stamp: 'cmm 7/25/2002 19:43'!
treeAt: aNode

	"Return a new instance of the receiver which has a single main
	parent aNode."

	^self
		treeAt: aNode
		ifAbsent: [ self error: aNode printString, ' was not found in ', self printString ]
! !

!MaTree methodsFor: 'accessing'!
treeAt: anObject ifAbsent: aBlock
	"Return a new instance of the receiver which has a single main
	parent anObject."

	| newTree |

	(self includes: anObject) ifFalse: [ ^aBlock value ].
	newTree := self class new.
	self
		depthFirstDo:
			[ :each :eachParent |
			newTree
				add: each
				asChildOf: eachParent ]
		at: anObject.
	^newTree
! !

!MaTree methodsFor: 'testing' stamp: 'cmm 7/25/2002 19:43'!
hasChildFor: anObject 
	"Returns true if anObject has any children.  Signal an error
	if anObject is not part of the hierarchy."

	^parents includes: anObject! !

!MaTree methodsFor: 'testing'!
hasParentFor: anObject
	"Returns true if anObject has a parent"

	^(self parentOf: anObject) notNil
! !

!MaTree methodsFor: 'testing'!
includes: anObject 
	^parents includesKey: anObject! !

!MaTree methodsFor: 'testing'!
pathTo: anObject includes: anotherObject
	"Returns a boolean.  Example:  in the Smalltalk
	class hierarchy, pathTo: Integer includes: Magnitude would
	return true.  We use pathFrom in the implementation because
	it is faster."

	^(self pathFrom: anObject) includes: anotherObject
! !

!MaTree methodsFor: 'private' stamp: 'cmm 1/23/2002 14:30'!
collect: aBlock into: aMaTree 
	"Since some objects may share a parent, we use the map to
	make sure we only value each parent just once, otherwise we
	could have some identity problems."

	| map |
	map := Dictionary new.
	parents keysAndValuesDo: 
			[:eachKey :eachValue | 
			eachValue isNil 
				ifTrue: [aMaTree add: (map at: eachKey ifAbsentPut: [aBlock value: eachKey])]
				ifFalse: 
					[aMaTree add: (map at: eachKey ifAbsentPut: [aBlock value: eachKey])
						asChildOf: (map at: eachValue ifAbsentPut: [aBlock value: eachValue])]].
	^aMaTree! !

!MaTree methodsFor: 'private' stamp: 'cmm 1/23/2002 14:30'!
initialize
	parents := Dictionary new! !

!MaTree methodsFor: 'printing'!
printOn: aStream

	super printOn: aStream.
	self depthFirstDo:
		[ :eachChild :eachParent |
		aStream
			cr;
			maTab: (self depthOf: eachChild);
			maPrintAbbreviated: eachChild ]
! !


!MaTree class methodsFor: 'creation'!
new

	^super new initialize
! !


!MaTree reorganize!
('building' add: add:asChildOf: addAll: addPath: addPath:to: addTree: addTree:to: copy move:toParent: remove: remove:ifAbsent:)
('enumerating' allLeafsDo: depthFirstDo: depthFirstDo:at: do: maDetect:ifFound:ifNone: nodesInPathFrom:to:do: pathFrom:do:)
('accessing' allLeafs allMembers children childrenOf: collect: conform: depthOf: detect: detect:ifNone: nodesInPathFrom:to: parentOf: pathFrom: pathTo: reject: roots select: size treeAt: treeAt:ifAbsent:)
('testing' hasChildFor: hasParentFor: includes: pathTo:includes:)
('private' collect:into: initialize)
('printing' printOn:)
!


!MaIntervalCollection reorganize!
('building' add: addFrom:to: remove: removeFrom:to:)
('private building' bridgeFrom:to: extendGapFromValue:at:toValue:at: extendIntervalFrom:at:to:at: privateInsertFrom:to:afterIndex: proximityThreshold: voidFrom:to:)
('private testing' fromValue:isOutsideProximityThresholdAt: isStartOfGap: isStartOfInterval: toValue:isOutsideProximityThresholdAt:)
('accessing' gapsDetect:ifFound:ifNone: highest indexOf: intervals maDetect:ifFound:ifNone: size)
('actions' gapsDo: intervalsDo:)
('testing' includes: isEmpty notEmpty)
('private' independentCopy initialize logicalIndexFor: physicalIndexFor: physicalIndexOf: physicalSize)
('printing' maPrintAbbreviatedOn:)
!

